(* file: rain.ml
 * author: Bob Muller
 * date: February 18, 2015
 * revised: September 30, 2016
 *
 * CS1103 Computer Science I Honors
 *
 * This programs simulates rain gathering in a box. Use the up and down arrow
 * keys to make it rain harder or not.
 *
 * depends: Asai/Uehara Universe Library
 * compile: > ./make
 * run:     > ./go
*)
open Color
open Image
open World
open Cs1103

(* constants
*)
let displayWidth  = 600.              (* window width *)
let displayHeight = displayWidth      (* window height *)
let boxDepth = 10.                    (* The 3D-box is the 2D-display
                                         with this depth *)
let pi = acos(-1.0)
let dropAmount = 5.0

type drop  = {x : float; y : float; radius : float}
type model = {drops : drop list; rate : int; volume : float}

(* volume : drop -> float
 *)
let dropVolume drop =
  (4.0 /. 3.0) *. pi *. (drop.radius ** 3.0)

(* boxVolume : float -> drop list -> float
 *)
let boxVolume formerVolume drops =
  let dropsVolume = List.fold_left (+.) 0.0 (List.map dropVolume drops)
  in
  formerVolume +. dropsVolume

(* waterLevel : volume -> float
 *)
let waterLevel volume = volume /. (displayWidth *. boxDepth)

(* makeDrop unit -> drop
 *)
let makeDrop () =
  {x = Random.float displayWidth;
   y = 0.0;
   radius = Random.float boxDepth }

(* makeDrops : int -> drop list
 *)
let rec makeDrops n =
  match n = 0 with
  | true  -> []
  | false -> makeDrop () :: makeDrops (n - 1)

(* fall : drop -> drop
 *)
let fall drop = { drop with y = drop.y +. dropAmount }

(* update : model -> model World.t
 *)
let update ({drops; rate; volume} as model) =
  let oldDrops = List.map fall drops in
  let level = waterLevel volume in
  let splashed drop = drop.y +. drop.radius >= (displayHeight -. level) in
  let splashed = List.filter splashed oldDrops in
  let stillFalling drop = drop.y +. drop.radius < (displayHeight -. level) in
  let stillFalling = List.filter stillFalling oldDrops in
  let newDrops = makeDrops (Random.int rate)
  in
  World { model with
          drops = stillFalling @ newDrops;
          volume = boxVolume volume splashed}

(* changeModelOnKey : model -> string -> model
 *)
let changeModelOnKey model name =
  match name with
  | "up"   -> World { model with rate = model.rate + 1}
  | "down" -> World { model with rate = model.rate - 1}
  | other  -> World model

(* placeDrop : drop -> Image.t -> Image.t
 *)
let placeDrop drop image =
  let (x, y) = (drop.x -. drop.radius, drop.y -. drop.radius)
  in
  Image.place_image (Image.circle drop.radius gray) (x, y) image

(* placeDrops : drop list -> Image.t -> Image.t
 *)
let rec placeDrops drops image =
  match drops with
  | [] -> image
  | drop :: drops -> placeDrops drops (placeDrop drop image)

(* view : model -> Image.t
 *)
let view {drops; rate; volume} =
  let empty = Image.empty_scene displayWidth displayHeight in
  let waterLevel = waterLevel volume in
  let accumulatedRain = rectangle displayWidth waterLevel blue in
  let scene = place_image accumulatedRain (0., displayHeight -. waterLevel) empty
  in
  placeDrops drops scene

let rain () =
  big_bang {drops = []; rate = 5; volume = 0.0}
           ~name: "Rain"
           ~to_draw: view
           ~width: (f2I displayWidth)
           ~height: (f2I displayHeight)
           ~on_tick: update
           ~on_key_press: changeModelOnKey
           ~rate: 0.05

let _ = rain ()
